#!/usr/env/bin python

import argparse
import enum
import functools
import importlib
import inspect
import logging
import multiprocessing
import os
import signal
import time
import typing
import types

from datetime import datetime

from bert import utils as bert_utils
from bert.deploy import utils as bert_deploy_utils

logger = logging.getLogger(__name__)

STOP_DAEMON: bool = False
class Service(enum.Enum):
    AWSLambda: str = 'aws-lambda'

def capture_options() -> typing.Any:
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--service', default=Service.AWSLambda, type=Service)
    parser.add_argument('-u', '--undeploy', action="store_true", default=False)
    parser.add_argument('-i', '--invoke', action="store_true", default=False)
    parser.add_argument('-m', '--module-name', default='bert')
    parser.add_argument('-f', '--flush', default=False, action="store_true")
    parser.add_argument('-d', '--dry-run', default=False, action="store_true", help="Create the lambda functions and output ./lambdas without deploying to AWS")
    return parser.parse_args()

def deploy_service(options) -> None:
    jobs: typing.Dict[str, typing.Any] = bert_utils.scan_jobs(options)
    jobs: typing.Dict[str, typing.Any] = bert_utils.map_jobs(jobs)

    if options.service == Service.AWSLambda:
        if options.invoke:
            import boto3
            job_name: str = [job for job in jobs.keys()][0]
            client = boto3.client('lambda')
            logger.info(f'Invoking Job[{job_name}]')
            client.invoke(
                FunctionName=job_name,
                InvocationType='Event')
            import sys; sys.exit(0)

        bert_deploy_utils.build_project(jobs)
        bert_deploy_utils.build_lambda_handlers(jobs)
        bert_deploy_utils.build_archives(jobs)
        bert_deploy_utils.create_roles(jobs)
        bert_deploy_utils.scan_dynamodb_tables(jobs)
        bert_deploy_utils.destroy_lambda_to_table_bindings(jobs)
        bert_deploy_utils.destroy_lambdas(jobs)
        bert_deploy_utils.create_lambdas(jobs)
        bert_deploy_utils.create_dynamodb_tables(jobs)
        bert_deploy_utils.bind_lambdas_to_tables(jobs)

    else:
        raise NotImplementedError(options.service)

def run_from_cli():
    import sys, os
    sys.path.append(os.getcwd())
    options = capture_options()
    deploy_service(options)

if __name__ in ['__main__']:
    run_from_cli()




import boto3
import collections
import glob
import logging
import inspect
import json
import os
import shutil
import tempfile
import time
import types
import typing
import zipfile

from bert import utils as bert_utils, encoders as bert_encoders, exceptions as bert_exceptions
from bert.deploy import shortcuts as bert_deploy_shortcuts, exceptions as deploy_exceptions
from botocore.errorfactory import ClientError

from distutils.sysconfig import get_python_lib

# Related: https://github.com/Miserlou/Zappa/pull/56
# Related: https://github.com/Miserlou/Zappa/pull/581
ZIP_EXCLUDES: typing.List[str] = [
    '*.exe', '*.DS_Store', '*.Python', '*.git', '.git/*', '*.zip', '*.tar.gz',
    '*.hg', 'pip', 'docutils*', 'setuputils*', '__pycache__/*',
]
COMMON_EXCLUDES: typing.List[str] = ['env', 'lambdas']
try:
    COMPRESSION_METHOD: int = zipfile.ZIP_DEFLATED
except ImportError: #pragma: no cover
    COMPRESSION_METHOD: int = zipfile.ZIP_STORED


logger = logging.getLogger(__name__)

def _calc_lambda_name(lambda_name: str) -> str:
    return lambda_name

def _calc_table_name(lambda_key: str) -> str:
    return lambda_key

def copytree(src: str, dest: str, metadata: bool = True, symlinks: bool = False, ignore: typing.Any = None) -> None:
    if isinstance(ignore, (types.GeneratorType, list, set, tuple)):
        ignore = shutil.ignore_patterns(*ignore)

    if not os.path.exists(dest):
        os.makedirs(dest)
        if metadata:
            shutil.copystat(src, dest)

    dir_list: typing.List[str] = os.listdir(src)
    if ignore:
        excl = ignore(src, dir_list)
        dir_list = [item for item in dir_list if item not in excl]

    for item in dir_list:
        source_path: str = os.path.join(src, item)
        dest_path: str = os.path.join(dest, item)
        if symlinks and os.path.islink(source_path):
            if os.path.lexsits(dest_path):
                os.remove(dest_path)

            os.symlink(os.readlink(source_path), dest_path)
            if metadata:
                try:
                    stats: typing.Any = os.lstat(source_path)
                    mode = stat.S_IMODE(stats.st_mode)
                    os.lchmod(dest_path, mode)
                except:
                    pass # lchmod not availabe

        elif os.path.isdir(source_path):
            copytree(source_path, dest_path, metadata, symlinks, ignore)

        else:
            shutil.copy2(source_path, dest_path) if metadata else shutil.copy(source_path, dest_path)

def find_site_packages_dir(start: str, find: str) -> str:
    """
    Scan a directory[start] for a subdirectory[find]
    """
    if not os.path.isdir(start):
        raise IOError(f'Path[{start}] is not a directory.')

    for root, dirs, files in os.walk(start):
        if find in dirs:
            return os.path.join(start, find)

        for dir_name in dirs:
            result: str = find_site_packages_dir(os.path.join(start, dir_name), find)
            if result:
                return result


def build_lambda_handlers(jobs: typing.Dict[str, typing.Dict[str, typing.Any]]) -> None:
    for job_name, conf in jobs.items():
        job_source: str = inspect.getsource(conf['job']).split('\n')
        job_templates: str = ''.join([f"""def {jn}():
    pass
""" for jn in jobs.keys() if jn != job_name])
        source_code: str = """
import typing
%s

from bert import utils, constants, binding, shortcuts, encoders

encoders.load_identity_encoders(%s)
encoders.load_queue_encoders(%s)
encoders.load_queue_decoders(%s)

%s

def %s_handler(event: typing.Dict[str, typing.Any] = {}, context: 'lambda_context' = None) -> None:

    records: typing.List[typing.Dict[str, typing.Any]] = event.get('Records', [])
    if len(records) > 0 and constants.DEBUG == False:
        constants.QueueType = constants.QueueTypes.StreamingQueue
        work_queue, done_queue, ologger = utils.comm_binders(%s)
        for record in records:
            if record['eventName'].lower() == 'INSERT'.lower():
                work_queue.local_put(record['dynamodb']['NewImage'])

    elif constants.DEBUG:
        constants.QueueType = constants.QueueTypes.LocalQueue
        work_queue, done_queue, ologger = utils.comm_binders(%s)
        work_queue.local_put(event)

    else:
        work_queue, done_queue, ologger = utils.comm_binders(%s)

    ologger.info(f'QueueType[{constants.QueueType}]')
    %s()

""" % (
    job_templates,
    conf['encoding']['identity_encoders'],
    conf['encoding']['queue_encoders'],
    conf['encoding']['queue_decoders'],
    '\n'.join(job_source),
    job_name,
    job_name,
    job_name,
    job_name,
    job_name)

        conf['lambda-path'] = os.path.join(conf['aws-build']['path'], f'{job_name}.py')
        logger.info(f'Creating Lambda Handler[{conf["lambda-path"]}]')
        with open(conf['lambda-path'], 'w') as stream:
            stream.write(source_code)


def build_package(job_name: str, job_conf: typing.Dict[str, typing.Any], excludes: typing.List[str] = COMMON_EXCLUDES) -> None:
    try:
        compression_method: int = zipfile.ZIP_DEFLATED
    except ImportError: #pragma: no cover
        compression_method: int = zipfile.ZIP_STORED

    archive_name: str = f'{job_name}.zip'
    archive_dir: str = os.path.join(os.getcwd(), 'lambdas')
    if not os.path.exists(archive_dir):
        os.makedirs(archive_dir)

    archive_path: str = os.path.join(archive_dir, archive_name)
    logger.info(f'Building Lambda[{job_name}] Archive[{archive_path}]')

    job_conf['archive-path'] = archive_path
    with zipfile.ZipFile(archive_path, 'w', compression_method) as archive:
        for root, dirs, files in os.walk(job_conf['project-path']):
            for filename in files:
                if filename in excludes:
                    continue

                if filename.endswith('.pyc'):
                    continue

                abs_filename: str = os.path.join(root, filename)
                if filename.endswith('.py'):
                    os.chmod(abs_filename, 0o755)

                zip_info: zipfile.ZipInfo = zipfile.ZipInfo(os.path.join(root.replace(job_conf['project-path'], '').lstrip(os.sep), filename))
                zip_info.create_system = 3
                zip_info.external_attr = 0o755 << int(16)
                with open(abs_filename, 'rb') as file_stream:
                    archive.writestr(zip_info, file_stream.read(), compression_method)

            for dirname in dirs:
                if dirname in excludes:
                    continue


    job_conf['archive-path'] = archive_path

def include_bert_dev(dev_path: str, build_path: str, excludes: typing.List[str] = []) -> None:
    # Make sure the correct filepath was provided
    for filename in ['__init__.py', 'runner', 'deploy', 'remote', 'binding.py', 'constants.py', 'utils.py', 'shortcuts.py']:
        assert filename in os.listdir(dev_path), f'Incorrect BERT_DEV[{bert_dev_path}] provided, filename[{filename}] not found'

    bert_path: str = os.path.join(build_path, 'bert')
    if not os.path.exists(bert_path):
        os.makedirs(bert_path)

    logger.warning('BERT_DEV ENVVar found, including development version of bert-etl')
    copytree(dev_path, bert_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))


def build_project(jobs: typing.Dict[str, typing.Any]) -> str:
    for job_name, conf in jobs.items():
        src_dir: str = None
        site_package_dir: str = None
        if not os.path.exists(conf['aws-build']['lambdas-path']):
            os.makedirs(conf['aws-build']['lambdas-path'])

        if os.path.exists(conf['aws-build']['path']):
            shutil.rmtree(conf['aws-build']['path'])

        os.makedirs(conf['aws-build']['path'])

        # Resolve venv site-packages locations
        if 'VIRTUAL_ENV' in os.environ.keys():
            site_package_dir: str = find_site_packages_dir(os.environ['VIRTUAL_ENV'], 'site-packages')
            src_dir: str = os.path.join(os.environ['VIRTUAL_ENV'], 'src')

        # Resolve conda env site-packages location
        elif 'conda/envs' in get_python_lib() or 'miniconda/envs' in get_python_lib():
            site_package_dir: str = get_python_lib()

        if os.path.exists('.python-version'):
            import ipdb; ipdb.set_trace()
            raise NotImplementedError

        # Resolve Egg-Links
        # https://setuptools.readthedocs.io/en/latest/formats.html#egg-links
        egg_links: typing.List[str] = []
        for root, dirnames, filenames in os.walk(site_package_dir):
            for filename in filenames:
                if filename.endswith('.egg-link'):
                    egg_links.append(os.path.join(site_package_dir, filename))

        for filepath in egg_links:
            with open(filepath, 'r') as stream:
                data: str = stream.read()
                basename: str = os.path.basename(data)
                if not '\n' in data:
                    logger.error(f'error parsing egg-link for base[{basename}]. Please open an issue request with this error')

                dirpath: str = data.split('\n')[0]
                logger.info(f'Including DevPackage-EggLink[{basename}]')
                copytree(dirpath, conf['aws-build']['path'], metadata=False, symlinks=False, ignore=conf['aws-build']['excludes'])

        # Include Dev Version of bert-etl if found
        if os.environ.get('BERT_DEV', None):
            include_bert_dev(os.environ['BERT_DEV'], conf['aws-build']['path'], conf['aws-build']['excludes'])

        logger.info(f'Copying PWD[{os.getcwd()}] to Build Dir[{conf["aws-build"]["path"]}]')
        copytree(os.getcwd(), conf['aws-build']['path'], metadata=False, symlinks=False, ignore=conf['aws-build']['excludes'])
        logger.info(f'Copying SitePackages[{site_package_dir}] to Build Dir[{conf["aws-build"]["path"]}]')
        copytree(site_package_dir, conf['aws-build']['path'], metadata=False, symlinks=False, ignore=conf['aws-build']['excludes'])
        if src_dir and os.path.exists(src_dir):
            logger.info(f'Merging SrcDir[{src_dir}]')
            copytree(src_dir, conf['aws-build']['path'], metadata=False, symlinks=False, ignore=conf['aws-build']['excludes'])

        if conf['aws-deploy']['requirements']:
            logger.info(f'Mergeing Job[{job_name}] requirements')
            bert_utils.run_command(f'pip install -t {conf["aws-build"]["path"]} {" ".join(conf["aws-deploy"]["requirements"])} -U')

def build_archives(jobs: typing.Dict[str, typing.Any]) -> str:
    for job_name, conf in jobs.items():
        logger.info(f'Building Lambda Archive[{conf["aws-build"]["archive-path"]}]')
        with zipfile.ZipFile(conf['aws-build']['archive-path'], 'w', COMPRESSION_METHOD) as archive:
            for root, dirs, files in os.walk(conf['aws-build']['path']):
                for filename in files:
                    if filename in conf['aws-build']['excludes']:
                        continue

                    if filename.endswith('.pyc'):
                        continue

                    abs_filepath: str = os.path.join(root, filename)
                    if filename.endswith('.py'):
                        os.chmod(abs_filepath, 0o755)

                    relative_filepath = os.path.join(root.replace(conf['aws-build']['path'], '').lstrip(os.sep), filename)
                    zip_info: zipfile.ZipInfo = zipfile.ZipInfo(relative_filepath)
                    zip_info.create_system = 3
                    zip_info.external_attr = 0o755 << int(16)
                    with open(abs_filepath, 'rb') as file_stream:
                        archive.writestr(zip_info, file_stream.read(), COMPRESSION_METHOD)

                for dirname in dirs:
                    if dirname in conf['aws-build']['excludes']:
                        continue

        conf['aws-build']['archive-size'] = os.path.getsize(conf['aws-build']['archive-path'])


def scan_dynamodb_tables(jobs: typing.Dict[str, typing.Dict[str, typing.Any]]) -> None:
    client = boto3.client('dynamodb')
    for job_name, conf in jobs.items():
        try:
            conf['aws-deployed']['work-table'] = client.describe_table(TableName=conf['aws-deploy']['work-table-name'])
        except ClientError as err:
            conf['aws-deployed']['work-table'] = None

        try:
            conf['aws-deployed']['done-table'] = client.describe_table(TableName=conf['aws-deploy']['done-table-name'])
        except ClientError as err:
            conf['aws-deployed']['done-table'] = None

def create_dynamodb_tables(jobs: typing.Dict[str, typing.Dict[str, typing.Any]]) -> None:
    client = boto3.client('dynamodb')
    for job_name, conf in jobs.items():
        try:
            conf['aws-deployed']['work-table'] = client.describe_table(TableName=conf['aws-deploy']['work-table-name'])
        except ClientError as err:
            logger.info(f'Creating Dynamodb Table[{conf["aws-deploy"]["work-table-name"]}]')
            client.create_table(
                    TableName=conf['aws-deploy']['work-table-name'],
                    KeySchema=[
                        {
                            'AttributeName': 'identity',
                            'KeyType': 'HASH'
                        }
                    ],
                    AttributeDefinitions=[
                        {
                            'AttributeName': 'identity',
                            'AttributeType': 'S'
                        }
                    ],
                    StreamSpecification={
                        'StreamEnabled': True,
                        'StreamViewType': 'NEW_IMAGE'
                    },
                    BillingMode='PAY_PER_REQUEST')

            conf['aws-deployed']['work-table'] = client.describe_table(TableName=conf['aws-deploy']['work-table-name'])

        try:
            conf['aws-deployed']['done-table'] = client.describe_table(TableName=conf['aws-deploy']['done-table-name'])
        except ClientError as err:
            logger.info(f'Creating Dynamodb Table[{conf["aws-deploy"]["done-table-name"]}]')
            client.create_table(
                    TableName=conf['aws-deploy']['done-table-name'],
                    KeySchema=[
                        {
                            'AttributeName': 'identity',
                            'KeyType': 'HASH'
                        }
                    ],
                    AttributeDefinitions=[
                        {
                            'AttributeName': 'identity',
                            'AttributeType': 'S'
                        }
                    ],
                    StreamSpecification={
                        'StreamEnabled': True,
                        'StreamViewType': 'NEW_IMAGE'
                    },
                    BillingMode='PAY_PER_REQUEST')

            conf['aws-deployed']['done-table'] = client.describe_table(TableName=conf['aws-deploy']['done-table-name'])

def create_roles(jobs: typing.Dict[str, typing.Any]) -> None:
    trust_policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": [
                        # "apigateway.amazonaws.com",
                        "lambda.amazonaws.com",
                        # "events.amazonaws.com",
                        # "dynamodb.amazonaws.com",
                    ]
                },
                "Action": "sts:AssumeRole",
            }
        ]
    }
    policy_document = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:*"
                ],
                "Resource": "arn:aws:dynamodb:*:*:table/bert-etl-*",
            },
            {
                "Effect": "Allow",
                "Action": "s3:*",
                "Resource": "*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                    "logs:CreateLogGroup",
                ],
                "Resource": "arn:aws:logs:*:*:*",
            }
        ]
    }
    trust_policy_name: str = 'bert-etl-lambda-execution-policy-trust'
    policy_name: str = 'bert-etl-lambda-execution-policy'
    iam_role = {
        'Path': '/',
        'RoleName': trust_policy_name,
        'AssumeRolePolicyDocument': json.dumps(trust_policy_document),
        'Description': 'Bert-ETL Lambda Execution Role',
    }
    iam_policy = {
        'Path': '/',
        'PolicyName': policy_name,
        'PolicyDocument': json.dumps(policy_document),
        'Description': 'Bert-ETL Lambda Execution Policy'
    }
    iam_client = boto3.client('iam')
    role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)
    if role is None:
        iam_client.create_role(**iam_role)
        role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)

    policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
    if policy is None:
        iam_client.create_policy(**iam_policy)
        policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
        iam_client.attach_role_policy(RoleName=trust_policy_name, PolicyArn=policy['Arn'])

    for job_name, conf in jobs.items():
        conf['aws-deployed']['iam-role'] = role
        conf['aws-deployed']['iam-policy'] = policy

def destroy_lambda_to_table_bindings(jobs: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for job_name, conf in jobs.items():
        if conf['aws-deployed']['work-table'] is None:
            continue

        logger.info(f'Destroying mappings between Lambda[{job_name}] and Work-Table[{conf["aws-deploy"]["work-table-name"]}]')
        for event_mapping in client.list_event_source_mappings(
                EventSourceArn=conf['aws-deployed']['work-table']['Table']['LatestStreamArn'],
                FunctionName=conf['aws-deploy']['lambda-name'])['EventSourceMappings']:
            client.delete_event_source_mapping(UUID=event_mapping['UUID'])

def destroy_lambdas(jobs: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for job_name, conf in jobs.items():
        try:
            client.delete_function(FunctionName=conf['aws-deploy']['lambda-name'])

        except ClientError as err:
            pass

        else:
            logger.info(f'Deleted Lambda[{job_name}]')

def create_lambda_s3_item(job_name: str, conf: typing.Dict[str, typing.Any]) -> None:
    if conf['deployment']['s3_bucket'] is None:
        raise bert_exceptions.BertConfigError('Archive over 50 MB. Please specify an s3_bucket to upload to. https://bert-etl.readthedocs.io/en/latest/bert-etl.yaml#s3_bucket')

    client = boto3.client('s3')

    try:
        client.head_bucket(Bucket=conf['deployment']['s3_bucket'])
    except ClientError:
        logger.info(f'Creating Bucket[{conf["deployment"]["s3_bucket"]}]')
        try:
            client.create_bucket(Bucket=conf['deployment']['s3_bucket'])
        except ClientError:
            raise NotImplementedError('''
Should not contain uppercase characters
- Should not contain underscores (_)
- Should be between 3 and 63 characters long
- Should not end with a dash
- Cannot contain two, adjacent periods
- Cannot contain dashes next to periods (e.g., "my-.bucket.com" and "my.-bucket" are invalid)
''')

    filename: str = os.path.basename(conf['aws-build']['archive-path'])
    logger.info(f'Uploading file[{filename}] to bucket[{conf["deployment"]["s3_bucket"]}]')
    client.upload_file(conf['aws-build']['archive-path'], conf['deployment']['s3_bucket'], filename)
    return {
        'S3Bucket': conf['deployment']['s3_bucket'],
        'S3Key': filename,
    }

def create_lambdas(jobs: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for job_name, conf in jobs.items():
        if conf['aws-build']['archive-size'] - 50000000 > 0:
            code_config = create_lambda_s3_item(job_name, conf)

        else:
            code_config = {
                'ZipFile': open(conf['aws-build']['archive-path'], 'rb').read()
            }


        try:
            client.get_function(FunctionName=conf['aws-deploy']['lambda-name'])['Configuration']
        except ClientError as err:
            logger.info(f'Creating AWSLambda for Job[{job_name}]')
            lambda_description = client.create_function(
                FunctionName=conf['aws-deploy']['lambda-name'],
                Runtime=conf['aws-deploy']['runtime'],
                MemorySize=conf['aws-deploy']['memory-size'],
                Role=conf['aws-deployed']['iam-role']['Arn'],
                Handler=conf['aws-deploy']['handler'],
                Code=code_config,
                Timeout=conf['aws-deploy']['timeout'],
                Environment={'Variables': conf['aws-deploy']['environment']})
            conf['aws-deployed']['aws-lambda'] = client.get_function(FunctionName=conf['aws-deploy']['lambda-name'])

        else:
            logger.info(f'Replacing AWSLambda for Job[{job_name}]')
            client.delete_function(FunctionName=conf['aws-deploy']['lambda-name'])
            client.create_function(
                FunctionName=conf['aws-deploy']['lambda-name'],
                Runtime=conf['aws-deploy']['runtime'],
                MemorySize=conf['aws-deploy']['memory-size'],
                Role=conf['aws-deploy']['iam-role']['Arn'],
                Handler=conf['aws-deploy']['handler'],
                Code=code_config,
                Timeout=conf['aws-deploy']['timeout'],
                Environment={'Variables': conf['aws-deploy']['environment']},
            )
            conf['aws-lambda'] = client.get_function(FunctionName=conf['aws-deploy']['lambda-name'])['Configuration']

def bind_lambdas_to_tables(jobs: typing.Dict[str, typing.Any]) -> None:
    client = boto3.client('lambda')
    for job_name, conf in jobs.items():
        logger.info(f'Mapping Lambda[{job_name}] to Work-Table[{conf["aws-deploy"]["work-table-name"]}]')
        client.create_event_source_mapping(
            EventSourceArn=conf['aws-deployed']['work-table']['Table']['LatestStreamArn'],
            FunctionName=conf['aws-deploy']['lambda-name'],
            Enabled=True,
            StartingPosition='LATEST')




import base64
import numpy as np

class NumpyIdentityEncoder(json.JSONEncoder):
    def default(self: PWN, obj: typing.Any) -> typing.Any:
        if isinstance(obj, (
            np.float32,
            np.float64,
            np.float16,
            np.complex32,
            np.complex64,
            np.int8,
            np.int16,
            np.int32,
            np.int64,
            np.uint8,
            np.uint16,
            np.uint32,
            np.uint64,
            np.intc,
            np.intp,
            np.bool_,
            np.ndarray)):
            return base64.b64encode(obj)

        return super(NumpyIdentityEncoder, self).default(obj)

def numpy_done_queue_encoder(obj: typing.Any) -> str:
    import ipdb; ipdb.set_trace()
    pass

def numpy_work_queue_decoder(obj: str) -> typing.Any:
    import ipdb; ipdb.set_trace()
    pass




import boto3
import logging
import os
import typing
import yaml

from bert.config import exceptions as config_exceptions

logger = logging.getLogger(__name__)

def get_if_exists(key: str, default: typing.Any, data_type: typing.Any, defaults: typing.Dict[str, str], env_vars: typing.Dict[str, str]) -> typing.Dict[str, str]:
    try:
        return data_type(env_vars[key])
    except KeyError:
        pass

    except ValueError:
        raise config_exceptions.BertConfigError(f'Key[{key}] is not DataType[{data_type}]')

    try:
        return data_type(defaults.get(key, default))
    except ValueError:
        raise config_exceptions.BertConfigError(f'Key[{key}] is not DataType[{data_type}]')


def merge_requirements(defaults: typing.List[str], requirements: typing.List[str]) -> typing.List[str]:
    merged: typing.List[str] = {value for value in defaults}
    for value in requirements:
        if value in merged:
            continue

        merged.append(value)

    return [item for item in merged]


def merge_env_vars(defaults: typing.Dict[str, str], env_vars: typing.Dict[str, str]) -> typing.Dict[str, str]:
    merged: typing.Dict[str, str] = {key: value for key, value in defaults.items()}
    for key, value in env_vars.items():
        merged[key] = value

    return merged


def merge_lists(main: typing.List[typing.Any], secondary: typing.List[typing.Any], defaults: typing.List[typing.Any]) -> typing.List[typing.Any]:
    if len(main) == 0 and len(secondary) == 0:
        return defaults

    # Preserve order or list
    merged = []
    for value in secondary:
        if value in merged:
            continue

        merged.append(value)

    for value in main:
        if value in merged:
            continue

        merged.append(value)

    return merged


def load_local_configuration() -> typing.Dict[str, typing.Any]:
    conf_path: str = os.path.join(os.getcwd(), 'bert-etl.yaml')
    if not os.path.exists(conf_path):
        logger.info("Configuration Path[{conf_path}] doesn't exist")
        return None

    with open(conf_path, 'r', encoding='utf-8') as stream:
        return yaml.load(stream.read(), Loader=yaml.FullLoader)




import collections
import getpass
import os
import redis
import typing

from urllib.parse import ParseResult, urlparse

PWN = typing.TypeVar('PWN')
class ENVVars:
    __slots__ = ('_env_vars', '_old_values')
    _env_vars: typing.Dict[str, str]
    _old_values: typing.Dict[str, str]

    def __init__(self: PWN, env_vars: typing.Dict[str, str]) -> None:
        self._env_vars = env_vars
        self._old_values = {}

    def __enter__(self: PWN) -> PWN:
        for key, value in self._env_vars.items():
            self._old_values[key] = os.environ.get(key, None)
            os.environ[key] = value

        return self

    def __exit__(self: PWN, type, value, traceback) -> None:
        for key, old_value in self._old_values.items():
            if old_value is None:
                del os.environ[key]

            else:
                os.environ[key] = old_value

        self._old_values = {}

class Postgres(collections.namedtuple('Postgres', ['host', 'port', 'dbname', 'username', 'password'])):
  __slots__ = ()
  @classmethod
  def ParseURL(cls: PWN, url: str) -> PWN:
    parts: ParseResult = urlparse(url)
    try:
      creds, rest = parts.netloc.split('@', 1)
    except ValueError:
      username = getpass.getuser()
      password = ''
      host, port = parts.netloc.split(':', 1)
    else:
      host, port = rest.split(':', 1)
      username, password = creds.split(':', 1)

    return cls(
      host=host,
      port=int(port),
      dbname=parts.path.strip('/'),
      username=username,
      password=password)

  def __enter__(self) -> PWN:
    os.environ['PGHOST'] = self.host
    os.environ['PGPORT'] = str(self.port)
    os.environ['PGDATABASE'] = self.dbname
    os.environ['PGUSER'] = self.username
    os.environ['PGPASSWORD'] = self.password

  def __exit__(self, type, value, traceback) -> None:
    del os.environ['PGHOST']
    del os.environ['PGPORT']
    del os.environ['PGDATABASE']
    del os.environ['PGUSER']
    del os.environ['PGPASSWORD']

class RedisConnection(collections.namedtuple('RedisConnection', ['host', 'port', 'db', 'client'])):
  __slots__ = ()
  @classmethod
  def ParseURL(cls: PWN, url: str) -> PWN:
    parts: ParseResult = urlparse(url)
    host, port = parts.netloc.split(':')
    return cls(
      host=host,
      port=int(port),
      db=parts.path.strip('/'),
      client=redis.Redis(host=host, port=int(port), db=parts.path.strip('/')))




import logging

logger = logging.getLogger(__name__)

class BertSettings(collections.NamedTuple):
    requirements: typing.List[str]
    identity_encoders: typing.List[typing.Any]
    queue_encoders: typing.List[typing.Any]
    queue_decoders: typing.List[typing.Any]
    env_vars: typing.Dict[str, str]
    runtime: str
    memory_size: int
    timeout: int

CONFIGS: typing.Dict[str, BertSettings] = {}
def capture(job_name: str) -> BertSettings:
    bert_configuration = bert_deploy_shortcuts.load_local_configuration()
    import ipdb; ipdb.set_trace()
    pass

# def build_project_envs(jobs: typing.Dict[str, types.FunctionType], venv_path: str, excludes: typing.List[str] = COMMON_EXCLUDES) -> typing.Dict[str, typing.Any]:
#     confs: typing.Dict[str, typing.Any] = {}
# 
#     for job_name, job in jobs.items():
#         identity_encoders: typing.List[str] = bert_deploy_shortcuts.merge_lists(
#             bert_configuration.get('every_lambda', {'identity_encoders': []}).get('identity_encoders', []),
#             bert_configuration.get(job_name, {'identity_encoders': []}).get('identity_encoders', []),
#             ['bert.encoders.base.IdentityEncoder'])
# 
#         queue_encoders: typing.List[str] = bert_deploy_shortcuts.merge_lists(
#             bert_configuration.get('every_lambda', {'queue_encoders': []}).get('queue_encoders', []),
#             bert_configuration.get(job_name,        {'queue_encoders': []}).get('queue_encoders', []),
#             ['bert.encoders.base.encode_aws_object'])
# 
#         queue_decoders: typing.List[str] = bert_deploy_shortcuts.merge_lists(
#             bert_configuration.get('every_lambda', {'queue_decoders': []}).get('queue_decoders', []),
#             bert_configuration.get(job_name,        {'queue_decoders': []}).get('queue_decoders', []),
#             ['bert.encoders.base.decode_aws_object'])
# 
#         bert_encoders.load_encoders_or_decoders(identity_encoders)
#         bert_encoders.load_encoders_or_decoders(queue_encoders)
#         bert_encoders.load_encoders_or_decoders(queue_decoders)
# 
#         # Align the lists with expectations of top -> bottom execution in bert-etl.yaml
#         # identity_encoders.reverse()
#         # queue_encoders.reverse()
#         # queue_decoders.reverse()
# 
#         runtime: int = bert_deploy_shortcuts.get_if_exists('runtime', 'python3.6', str, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
#         memory_size: int = bert_deploy_shortcuts.get_if_exists('memory_size', '128', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
#         if int(memory_size / 64) != memory_size / 64:
#             raise deploy_exceptions.BertConfigError(f'MemorySize[{memory_size}] must be a multiple of 64')
# 
#         timeout: int = bert_deploy_shortcuts.get_if_exists('timeout', '900', int, bert_configuration.get('every_lambda', {}), bert_configuration.get(job_name, {}))
#         env_vars: typing.Dict[str, str] = bert_deploy_shortcuts.merge_env_vars(
#             bert_configuration.get('every_lambda', {'environment': {}}).get('environment', {}),
#             bert_configuration.get(job_name, {'environment': {}}).get('environment', {}))
# 
#         requirements: typing.Dict[str, str] = bert_deploy_shortcuts.merge_requirements(
#             bert_configuration.get('every_lambda', {'requirements': {}}).get('requirements', {}),
#             bert_configuration.get(job_name, {'requirements': {}}).get('requirements', {}))
# 
#         env_vars['BERT_QUEUE_TYPE'] = env_vars.get('BERT_QUEUE_TYPE', 'dynamodb')
#         project_path: str = tempfile.mkdtemp(prefix=f'bert-etl-project-{job_name}')
#         logger.info(f'Creating Project Path[{project_path}]')
#         job_templates: str = ''.join([f"""def {jn}():
#     pass
# """ for jn in jobs.keys() if jn != job_name])
# 
#         job_follow: str = inspect.getsource(job).split('\n')[0]
#         job_source: str = '\n'.join(inspect.getsource(job).split('\n')[2:])
#         job_template: str = """
# %s
# import typing
# 
# from bert import utils, constants, binding, shortcuts, encoders
# encoders.load_identity_encoders(%s)
# encoders.load_queue_encoders(%s)
# encoders.load_queue_decoders(%s)
# 
# %s
# def %s(event: typing.Dict[str, typing.Any] = {}, context: 'lambda_context' = None) -> None:
# 
#     records: typing.List[typing.Dict[str, typing.Any]] = event.get('Records', [])
#     if len(records) > 0 and constants.DEBUG == False:
#         constants.QueueType = constants.QueueTypes.StreamingQueue
#         work_queue, done_queue, ologger = utils.comm_binders(%s)
#         for record in records:
#             if record['eventName'].lower() == 'INSERT'.lower():
#                 work_queue.local_put(record['dynamodb']['NewImage'])
# 
#     elif constants.DEBUG:
#         constants.QueueType = constants.QueueTypes.LocalQueue
#         work_queue, done_queue, ologger = utils.comm_binders(%s)
#         work_queue.local_put(event)
# 
#     else:
#         work_queue, done_queue, ologger = utils.comm_binders(%s)
# 
#     ologger.info(f'QueueType[{constants.QueueType}]')
# 
# %s
#     return {}
# """ % (job_templates, identity_encoders, queue_encoders, queue_decoders, job_follow, job_name, job_name, job_name, job_name, job_source)
#         job_path: str = os.path.join(project_path, f'{job_name}.py')
# 
#         with open(job_path, 'w') as stream:
#             stream.write(job_template)
# 
#         logger.info(f'Creating Job[{job_name}] Project')
#         copytree(os.getcwd(), project_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
#         logger.info(f'Merging Job[{job_name}] Site Packages')
#         copytree(venv_path, project_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
#         logger.info(f'Merging Job[{job_name}] Requirements')
#         bert_utils.run_command(f'pip install -t {project_path} {" ".join(requirements)} -U')
#         confs[_calc_lambda_name(job_name)] = {
#                 'project-path': project_path,
#                 'table-name': f'{job_name}-stream',
#                 'timeout': timeout,
#                 'runtime': runtime,
#                 'memory-size': memory_size, # must be a multiple of 64, increasing memory size also increases cpu allocation
#                 'environment': env_vars,
#                 'requirements': requirements,
#                 'handler-name': f'{job_name}.{job_name}',
#                 'spaces': {
#                     'work-key': job.work_key,
#                     'done-key': job.done_key,
#                     'pipeline-type': job.pipeline_type,
#                     'workers': job.workers,
#                     'scheme': job.schema,
#                 },
#                 'encoding': {
#                     'identity_encoders': identity_encoders,
#                     'queue_encoders': queue_encoders,
#                     'queue_decoders': queue_decoders,
#                 }
#             }
# 
#     return confs
# 
# def build_package(job_name: str, job_conf: typing.Dict[str, typing.Any], excludes: typing.List[str] = COMMON_EXCLUDES) -> None:
#     try:
#         compression_method: int = zipfile.ZIP_DEFLATED
#     except ImportError: #pragma: no cover
#         compression_method: int = zipfile.ZIP_STORED
# 
#     archive_name: str = f'{job_name}.zip'
#     archive_dir: str = os.path.join(os.getcwd(), 'lambdas')
#     if not os.path.exists(archive_dir):
#         os.makedirs(archive_dir)
# 
#     archive_path: str = os.path.join(archive_dir, archive_name)
#     logger.info(f'Building Lambda[{job_name}] Archive[{archive_path}]')
# 
#     job_conf['archive-path'] = archive_path
#     with zipfile.ZipFile(archive_path, 'w', compression_method) as archive:
#         for root, dirs, files in os.walk(job_conf['project-path']):
#             for filename in files:
#                 if filename in excludes:
#                     continue
# 
#                 if filename.endswith('.pyc'):
#                     continue
# 
#                 abs_filename: str = os.path.join(root, filename)
#                 if filename.endswith('.py'):
#                     os.chmod(abs_filename, 0o755)
# 
#                 zip_info: zipfile.ZipInfo = zipfile.ZipInfo(os.path.join(root.replace(job_conf['project-path'], '').lstrip(os.sep), filename))
#                 zip_info.create_system = 3
#                 zip_info.external_attr = 0o755 << int(16)
#                 with open(abs_filename, 'rb') as file_stream:
#                     archive.writestr(zip_info, file_stream.read(), compression_method)
# 
#             for dirname in dirs:
#                 if dirname in excludes:
#                     continue
# 
# 
#     job_conf['archive-path'] = archive_path
# 
# def include_bert_dev(bert_dev_path: str, venv_path: str, excludes: typing.List[str] = []) -> None:
#     excludes = ZIP_EXCLUDES + excludes + ['lamdbas']
#     temp_bert_path: str = tempfile.mkdtemp(prefix='bert-dev-path')
#     bert_dev: str = os.path.join(temp_bert_path, 'bert')
#     # Make sure the correct filepath was provided
#     for filename in ['__init__.py', 'factory.py', 'binding.py', 'constants.py', 'utils.py', 'shortcuts.py']:
#         assert filename in os.listdir(bert_dev_path), f'Incorrect BERT_DEV[{bert_dev_path}] provided, filename[{filename}] not found'
# 
#     copytree(bert_dev_path, bert_dev, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
#     copytree(temp_bert_path, venv_path, metadata=False, symlinks=False, ignore=shutil.ignore_patterns(*excludes))
# 
# 
# def build_lambda_archives(jobs: typing.Dict[str, types.FunctionType]) -> str:
#     venv_path: str = get_current_venv()
#     if venv_path is None:
#         venv_path = get_pyenv_venv()
#         if venv_path is None:
#             venv_path = get_conda_venv()
#             if venv_path is None:
#                 raise NotImplementedError
# 
#     if os.environ.get('BERT_DEV', None):
#         logger.warning('BERT_DEV ENVVar found, including development version of bert-etl')
#         include_bert_dev(os.environ['BERT_DEV'], venv_path)
# 
#     archive_paths: typing.List[str] = []
#     lambdas: typing.Dict[str, typing.Any] = collections.OrderedDict()
#     for job_name, job_conf in build_project_envs(jobs, venv_path).items():
#         build_package(job_name, job_conf)
#         lambdas[job_name] = job_conf
#         # shutil.rmtree(venv_path)
#         shutil.rmtree(job_conf['project-path'])
# 
#     return lambdas
# 
# # def replace_lambda_archives_with_requires(build_lambda_archives
# def destroy_dynamodb_tables(jobs: typing.Dict[str, typing.Any]) -> None:
#     table: typing.Any = None
#     client = boto3.client('dynamodb')
#     for job_name, job in jobs.items():
#         work_table_name: str = _calc_table_name(job.work_key)
#         done_table_name: str = _calc_table_name(job.done_key)
# 
#         try:
#             client.delete_table(TableName=work_table_name)
#         except ClientError as err:
#             pass
# 
#         else:
#             logger.info(f'Destorying Table[{work_table_name}]')
# 
#         try:
#             client.delete_table(TableName=done_table_name)
#         except ClientError as err:
#             pass
# 
#         else:
#             logger.info(f'Destorying Table[{done_table_name}]')
# 
# 
# def build_dynamodb_tables(lambdas: typing.Dict[str, typing.Any]) -> None:
#     table: typing.Any = None
#     client = boto3.client('dynamodb')
#     for lambda_name, conf in lambdas.items():
#         work_table_name: str = _calc_table_name(conf['spaces']['work-key'])
#         done_table_name: str = _calc_table_name(conf["spaces"]["done-key"])
# 
#         try:
#             conf['work-table'] = client.describe_table(TableName=work_table_name)
#         except ClientError as err:
#             logger.info(f'Creating Dynamodb Table[{work_table_name}]')
#             client.create_table(
#                     TableName=work_table_name,
#                     KeySchema=[
#                         {
#                             'AttributeName': 'identity',
#                             'KeyType': 'HASH'
#                         }
#                     ],
#                     AttributeDefinitions=[
#                         {
#                             'AttributeName': 'identity',
#                             'AttributeType': 'S'
#                         }
#                     ],
#                     StreamSpecification={
#                         'StreamEnabled': True,
#                         'StreamViewType': 'NEW_IMAGE'
#                     },
#                     BillingMode='PAY_PER_REQUEST')
#             conf['work-table'] = client.describe_table(TableName=work_table_name)
# 
#         try:
#             conf['done-table'] = client.describe_table(TableName=done_table_name)
#         except ClientError as err:
#             logger.info(f'Creating Dynamodb Table[{done_table_name}]')
#             client.create_table(
#                     TableName=done_table_name,
#                     KeySchema=[
#                         {
#                             'AttributeName': 'identity',
#                             'KeyType': 'HASH'
#                         }
#                     ],
#                     AttributeDefinitions=[
#                         {
#                             'AttributeName': 'identity',
#                             'AttributeType': 'S'
#                         }
#                     ],
#                     StreamSpecification={
#                         'StreamEnabled': True,
#                         'StreamViewType': 'NEW_IMAGE'
#                     },
#                     BillingMode='PAY_PER_REQUEST')
# 
#             conf['done-table'] = client.describe_table(TableName=done_table_name)
# 
# 
# def create_lambda_roles(lambdas: typing.Dict[str, typing.Any]) -> None:
#     trust_policy_document = {
#         "Version": "2012-10-17",
#         "Statement": [
#             {
#                 "Sid": "",
#                 "Effect": "Allow",
#                 "Principal": {
#                     "Service": [
#                         "apigateway.amazonaws.com",
#                         "lambda.amazonaws.com",
#                         "events.amazonaws.com",
#                         "dynamodb.amazonaws.com",
#                     ]
#                 },
#                 "Action": "sts:AssumeRole",
#             }
#         ]
#     }
# 
#     policy_document: typing.Dict[str, typing.Any] = {
#         "Version": "2012-10-17",
#         "Statement": [
#             # {
#             #     "Effect": "Allow",
#             #     "Action": [
#             #         "xray:PutTraceSegments",
#             #         "xray:PutTelemetryRecords",
#             #     ],
#             #     "Resource": ["*"],
#             # },
#             # {
#             #     "Effect": "Allow",
#             #     "Action": [
#             #         "lambda:InvokeFunction"
#             #     ],
#             #     "Resource": ["*"],
#             # },
#             {
#                 "Effect": "Allow",
#                 "Action": [
#                     "dynamodb:*",
#                 ],
#                 "Resource": "arn:aws:dynamodb:*:*:table/*",
#             },
#             {
#                 "Effect": "Allow",
#                 "Action": "s3:*",
#                 "Resource": "*"
#             },
#             {
#                 "Effect": "Allow",
#                 "Action": [
#                     "logs:CreateLogStream",
#                     "logs:PutLogEvents",
#                     "logs:CreateLogGroup",
#                 ],
#                 "Resource": "arn:aws:logs:*:*:*",
#             }
#         ]
#     }
#     trust_policy_name: str = 'bert-etl-lambda-execution-policy-trust'
#     policy_name: str = 'bert-etl-lambda-execution-policy'
#     iam_role = {
#         'Path': '/',
#         'RoleName': trust_policy_name,
#         'AssumeRolePolicyDocument': json.dumps(trust_policy_document),
#         'Description': 'Bert-ETL Lambda Execution Role',
#     }
#     iam_policy = {
#         'Path': '/',
#         'PolicyName': policy_name,
#         'PolicyDocument': json.dumps(policy_document),
#         'Description': 'Bert-ETL Lambda Execution Policy'
#     }
#     iam_client = boto3.client('iam')
#     role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)
#     if role is None:
#         iam_client.create_role(**iam_role)
#         role = bert_deploy_shortcuts.map_iam_role(trust_policy_name)
# 
#     policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
#     if policy is None:
#         iam_client.create_policy(**iam_policy)
#         policy = bert_deploy_shortcuts.map_iam_policy(policy_name)
#         iam_client.attach_role_policy(RoleName=trust_policy_name, PolicyArn=policy['Arn'])
# 
#     for job_name, conf in lambdas.items():
#         conf['iam-role'] = role
#         conf['iam-policy'] = policy
# 
# def destory_lambda_to_table_bindings(lambdas: typing.Dict[str, typing.Any]) -> None:
#     client = boto3.client('lambda')
#     for lambda_name, conf in lambdas.items():
#         for event_mapping in client.list_event_source_mappings(
#                 EventSourceArn=conf['work-table']['Table']['LatestStreamArn'],
#                 FunctionName=lambda_name)['EventSourceMappings']:
#             client.delete_event_source_mapping(UUID=event_mapping['UUID'])
# 
# def destroy_lambdas(jobs: typing.Dict[str, typing.Any]) -> None:
#     client = boto3.client('lambda')
#     for job_name, job in jobs.items():
#         lambda_name: str = _calc_lambda_name(job_name)
#         try:
#             client.delete_function(FunctionName=lambda_name)
#         except ClientError as err:
#             pass
# 
#         else:
#             logger.info(f'Deleting function[{lambda_name}]')
# 
# 
# def upload_lambdas(lambdas: typing.Dict[str, typing.Any]) -> None:
#     client = boto3.client('lambda')
#     for lambda_name, conf in lambdas.items():
#         try:
#             client.get_function(FunctionName=lambda_name)['Configuration']
#         except ClientError as err:
#             logger.info(f'Creating AWSLambda for Job[{lambda_name}]')
#             lambda_description = client.create_function(
#                 FunctionName=lambda_name,
#                 Runtime=conf['runtime'],
#                 MemorySize=conf['memory-size'],
#                 Role=conf['iam-role']['Arn'],
#                 Handler=conf['handler-name'],
#                 Code={
#                     'ZipFile': open(conf['archive-path'], 'rb').read(),
#                 },
#                 Timeout=conf['timeout'],
#                 Environment={'Variables': conf['environment']},
#             )
#             conf['aws-lambda'] = client.get_function(FunctionName=lambda_name)['Configuration']
# 
#         else:
#             logger.info(f'Replacing AWSLambda for Job[{lambda_name}]')
#             client.delete_function(FunctionName=lambda_name)
#             client.create_function(
#                 FunctionName=lambda_name,
#                 Runtime=conf['runtime'],
#                 MemorySize=conf['memory-size'],
#                 Role=conf['iam-role']['Arn'],
#                 Handler=conf['handler-name'],
#                 Code={
#                     'ZipFile': open(conf['archive-path'], 'rb').read(),
#                 },
#                 Timeout=conf['timeout'],
#                 Environment={'Variables': conf['environment']},
#             )
#             conf['aws-lambda'] = client.get_function(FunctionName=lambda_name)['Configuration']
# 
# def bind_lambdas_to_tables(lambdas: typing.Dict[str, typing.Any]) -> None:
#     client = boto3.client('lambda')
#     for lambda_name, conf in lambdas.items():
#         logger.info(f'Mapping Lambda[{lambda_name}] to Work-Table')
#         client.create_event_source_mapping(
#                 EventSourceArn=conf['work-table']['Table']['LatestStreamArn'],
#                 FunctionName=lambda_name,
#                 Enabled=True,
#                 StartingPosition='LATEST')
# 
